/*
 * Created on Jul 6, 2005
 *
 */
package tacle.core;

import java.util.Set;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;

/**
 * Generic interface to a call graph construction analysis.
 */

public interface ICallGraphAnalysis {
    
    /** 
     * Run the analysis, starting from the given main method.
     */
    void run(IMethod mainMethod);
    
    /**
     * Returns a set of <code>IMethod</code> references, representing
     * the methods reachable from the main method. Not guarantted to
     * be complete, due to the limitations of <code>IType</code> and
     * <code>IMethod</code>. Static initializers are not represented 
     * by IMethods, and any calls from them are not represented.
     * 
     * The current version does not deal with reflective calls and throw 
     * clauses. Methods declared in anonymous classes are not reported.
     * The body of a method declared in an anonymous class created that
     * is declared in a method may be wrongly reported as belonging
     * to the enclosing method. See 
     * http://presto.cse.ohio-state.edu/tacle/index.html
     * for more details on the current implementation.   
     * 
     * Current version works only with features of Java 1.4.
     */
    Set getReachableIMethods();

    /**
     * Returns a set of <code>IType</code> references, representing
     * the class and interface types that are initialized in code
     * reachable from the main method. Not guarantted to be complete,
     * due to the limitations of <code>IType</code> and
     * <code>IMethod</code>. 
     * 
     * See 
     * http://presto.cse.ohio-state.edu/tacle/index.html
     * for more details on the current implementation. 
     */
    Set getInitializedITypes();

    /** 
     * Returns a set of <code>IMethod</code> references for all
     * methods called by the given method. The given method should be
     * reachable. Not guarantted to be complete, due to the
     * limitations of <code>IType</code> and
     * <code>IMethod</code>.
     *  
     * The current version does not deal with reflective calls and throw 
     * clauses. Methods declared in anonymous classes are not reported.
     * The body of a method declared in an anonymous class created that
     * is declared in a method may be wrongly reported as belonging
     * to the enclosing method. See 
     * http://presto.cse.ohio-state.edu/tacle/index.html
     * for more details on the current implementation.   
     * 
     * Current version works only with features of Java 1.4.
     */    
    Set getCalledIMethods(IMethod method);

    /** 
     * Returns the set of call sites in the method. The given method
     * should be reachable. 
     * @see ICallSite
     */    
    Set getCallSites(IMethod method);

    /** 
     * Reports true the given method was analyzed from source code, and false
     * otherwise. The given method should be reachable.
     */
    boolean analyzedFromSourceCode(IMethod method);

    /** 
     * Looks for a call site in the given method such that the start
     * position and the length of the source range match exactly the
     * given values. This method can be used to obtain information
     * about an AST node representing a call site: 
     * <blockquote> <pre>
     * IMethod m = ...;
     * MethodInvocation n = ...; // node in m's AST
     * CallGraphAnalysis c = ...;
     * CallSite s = c.getCallSite(m, n.getStartPosition(), n.getLength());
     * if (s != null) // null will be returned if there is no match
     *    System.out.println("Num callees: " + s.getCalledIMethods().size());
     * </pre> </blockquote>
     * The given method should be reachable. 
     */
    ICallSite getCallSite(IMethod method, int pos, int length);

    /** 
     * Returns a set of <code>IMethod</code> references for all
     * methods calling the given method. The given method should be
     * reachable. Not guarantted to be complete, due to the
     * limitations of <code>IType</code> and
     * <code>IMethod</code>.
     *  
     * The current version does not deal with reflective calls and throw 
     * clauses. Methods declared in anonymous classes are not reported.
     * The body of a method declared in an anonymous class created that
     * is declared in a method may be wrongly reported as belonging
     * to the enclosing method. See 
     * http://presto.cse.ohio-state.edu/tacle/index.html
     * for more details on the current implementation.   
     * 
     * Current version works only with features of Java 1.4.
     */
    public Set getCallingMethods(IMethod method);

    /**
     * Adds a method declared in the library 
     * to the list of reachable methods.
     * The elements of the array in the parameter <code>method</code> 
     * represent, in this order: <br>
     * - package name of class declaring the method <br> 
     * - complete name of class declaring the method <br>
     * - package name of return type (or "" if none) <br>
     * - complete name of return type <br>
     * - name of method <br>
     * - package name of class of first parameter (or "" if none) <br>
     * - complete name of type of first parameter <br>
     * - ... <br>
     * (list continues for the other parameters). Package names
     * are provided separately in order to avoid confusions 
     * in finding the package name for 
     * anonymous and inner types.
     * 
     * This method should be called before the method
     * <code>run</code> is called in order to have effect in the 
     * analysis. 
     */
    public void addLibraryReachableMethod(String[] method);

    /**
     * Adds the method declared in user code <code>method</code>
     * to the list of reachable methods.
     * This method should be called before the method
     * <code>run</code> is called in order to have effect in the 
     * analysis. 
     */
    public void addReachableMethod(IMethod method);

    /**
     * Adds the library class <code>className</code> to the list of
     * instantiated classes, if it is not already there.
     * <code>className</code> is the complete name of the class, 
     * including the package name. <code>packageName</code> is a substring
     * of <code>className</code> (or "") and is provided in order
     * to avoid confusions in finding the package name for 
     * anonymous and inner types.
     * 
     * This method should be called before the method
     * <code>run</code> is called in order to have effect in the 
     * analysis. 
     */
    public void addLibraryInstantiatedClass(String packageName, String className);

    /**
     * Adds the type declared in user code <code>type</code> to the list of
     * instantiated classes, if it is not already there. 
     * This method should be called before the method
     * <code>run</code> is called in order to have effect in the 
     * analysis. 
     */
    public void addInstantiatedClass(IType type);
}
    
