/*
 * Created on Aug 5, 2005
 *
 */
package tacle.internal.rta;

import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.ConstructorInvocation;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.Initializer;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.StringLiteral;
import org.eclipse.jdt.core.dom.SuperConstructorInvocation;
import org.eclipse.jdt.core.dom.SuperMethodInvocation;
import org.eclipse.jdt.core.dom.TypeDeclaration;

import tacle.internal.rta.data.IWorklistType;
import tacle.internal.rta.data.IWorklistMethod;
import tacle.internal.rta.data.InfoDictionary;

/**
 * @author Mariana L Barca
 * 
 * visits reachable code in constructor and finalizer (if any) if a finalizer
 * exists, it is reported by the method hasFinalizer() this should be used to
 * add finalizers to the reachable methods AFTER they are visited (should not be
 * added to worklist)
 */
public class ConstructorVisitor extends MethodVisitor {

	private boolean isInConstructor;

	private boolean isInClass;

	private IWorklistType[] params;

	private IWorklistType cls;

	private String parentDefContrKey;

	private IWorklistType parent;

	// fields that report meaningful information
	// only after a traversal
	private boolean hasCalledSuperThis;

	private boolean hasFinalize;

	private String finalizerKey;

	// getters
	public String getFinalizerKey() {
		return finalizerKey;
	}

	public IWorklistType getParent() {
		return parent;
	}

	public String getParentDefaultConstrKey() {
		return parentDefContrKey;
	}

	// to be called only after visit
	// needed to remember the call to other constructors (super or this)
	public boolean hasCalledAnotherConstructor() {
		return hasCalledSuperThis;
	}

	// true if the class has a finalizer
	public boolean hasFinalizer() {
		return hasFinalize;
	}

	// need to know which constructor to analyze
	ConstructorVisitor(RTAMethodWorklist worklist, IWorklistType cls,
			IWorklistMethod method) {
		super(worklist, method);
		isInConstructor = false;
		isInClass = false;
		hasCalledSuperThis = false;
		this.params = method.getParameters();
		this.cls = cls;

		parentDefContrKey = null;
		parent = null;
	}

	//// visiting methods

	public boolean visit(ClassInstanceCreation node) {
		if (isInConstructor) {
			//System.out.println("Visiting... " + node);
			super.visit(node);
		}
		return true;
	}

	public boolean visit(ConstructorInvocation node) {
		if (isInConstructor) {
			//System.out.println("Visiting... " + node);
			hasCalledSuperThis = true;
			super.visit(node);
		}
		return true;
	}

	// visit only instance field declarations
	public boolean visit(FieldDeclaration node) {
		if (!Modifier.isStatic(node.getModifiers())) {
			isInConstructor = true;
		} else {
			isInConstructor = false;
		}
		return isInConstructor;
	}

	// assumed that the flags say if this is a static block
	public boolean visit(Initializer node) {
		//System.out.println(node);
		if (isInClass && !Modifier.isStatic(node.getModifiers())) {
			isInConstructor = true;
		} else {
			isInConstructor = false;
		}
		return isInConstructor;
	}

	public boolean visit(MethodDeclaration node) {
		IMethodBinding mb = node.resolveBinding();
		if (isInClass) {
			if (node.isConstructor()) {
				//System.out.println("Visiting... " + node);
				// check that this is the constructor we are looking for
				// and not the one of a different class declared in the same cu
				IWorklistType temp = InfoDictionary.v().getTypeForBinding(
						mb.getDeclaringClass());
				if (cls.equals(temp)
						&& params.length == mb.getParameterTypes().length) {
					isInConstructor = true;
					for (int j = 0; j < params.length; j++) {
						temp = InfoDictionary.v().getTypeForBinding(
								mb.getParameterTypes()[j]);
						if (!params[j].equals(temp))
							isInConstructor = false;
					}
				} else {
					isInConstructor = false;
				}
			} else {
				if (mb.getName().equals("finalize")) {
					hasFinalize = true;
					finalizerKey = mb.getKey();
				}
				isInConstructor = false;
			}
			return isInConstructor;
		} else {
			return true; // may be in surrounding type/method
		}
	}

	public boolean visit(MethodInvocation node) {
		if (isInConstructor) {
			//System.out.println(method + "Visiting... " + node);
			super.visit(node);
		}
		return true;
	}

	// this can be the name of a static field
	public boolean visit(SimpleName node) {
		if (isInConstructor) {
			super.visit(node);
		}
		return true;
	}

	public boolean visit(StringLiteral node) {
		if (isInConstructor) {
			super.visit(node);
		}
		return true;
	}

	public boolean visit(SuperConstructorInvocation node) {
		if (isInConstructor) {
			//System.out.println("Visiting... " + node);
			hasCalledSuperThis = true;
			super.visit(node);
		}
		return true;
	}

	public boolean visit(SuperMethodInvocation node) {
		if (isInConstructor) {
			//System.out.println("Visiting... " + node);
			super.visit(node);
		}
		return true;
	}

	// this is to remember the parent
	// in case super was not called explicitly and
	// we have to call it manually
	public boolean visit(TypeDeclaration node) {
		ITypeBinding itb = node.resolveBinding();
		IWorklistType temp = InfoDictionary.v().getTypeForBinding(itb);
		if (cls.equals(temp)) {
			isInClass = true;
			if (itb.getSuperclass() != null) {
				parent = InfoDictionary.v().getTypeForBinding(
						itb.getSuperclass());
				// search for the key of the default constructor
				IMethodBinding[] methods = itb.getSuperclass()
						.getDeclaredMethods();
				// which one is the default constr
				for (int i = 0; i < methods.length; i++) {
					if (methods[i].isDefaultConstructor()) {
						parentDefContrKey = methods[i].getKey();
					}
				}
			}
		}
		return true;
	}

	// end visit
	
	// this is to remember the parent
	// in case super was not called explicitly and
	// we have to call it manually
	public void endVisit(TypeDeclaration node) {
		ITypeBinding itb = node.resolveBinding();
		IWorklistType temp = InfoDictionary.v().getTypeForBinding(itb);
		if (cls.equals(temp)) {
			isInClass = false;
		}
	}

}
