/*
 * Created on Aug 5, 2005
 *
 */
package tacle.internal.rta;

import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.ConstructorInvocation;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.Initializer;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.TypeDeclaration;
//import org.eclipse.jdt.core.dom.StringLiteral;
import org.eclipse.jdt.core.dom.SuperConstructorInvocation;
import org.eclipse.jdt.core.dom.SuperMethodInvocation;

import tacle.internal.rta.data.IWorklistType;
import tacle.internal.rta.data.IWorklistMethod;
import tacle.internal.rta.data.InfoDictionary;

/**
 * @author Mariana L Barca
 * 
 * filters execution of method visitor to apply only to aspects relevant to
 * static initializers the whole compilation unit must be visited, unlike
 * methodVisitor which is applied only to the AST of a method in case of a
 * regular method (not constructor, not initializer)
 */
public class StaticInitializerVisitor extends MethodVisitor {

	// is in initializer code
	private boolean isInInitializer;

	// visitor is inside the class we are looking for
	private boolean isInsideClass;

	// the class for which we analyze the static initializer
	private IWorklistType cls;

	// field relevant only after traversal
	private boolean hasInitializer;

	public boolean hasStaticInitializer() {
		return hasInitializer;
	}

	// need to know which constructor to analyze
	StaticInitializerVisitor(RTAMethodWorklist worklist, IWorklistType cls, IWorklistMethod method) {
		super(worklist, method);
		isInInitializer = false;
		isInsideClass = false;
		hasInitializer = false;
		this.cls = cls;
	}

	public boolean visit(ClassInstanceCreation node) {
		if (isInInitializer) {
			//System.out.println("Visiting... " + node);
			super.visit(node);
		}
		return true;
	}

	public boolean visit(ConstructorInvocation node) {
		if (isInInitializer) {
			//System.out.println("Visiting... " + node);
			super.visit(node);
		}
		return true;
	}

	// visit only static field declarations
	public boolean visit(FieldDeclaration node) {
		if (isInsideClass && Modifier.isStatic(node.getModifiers())) {
			isInInitializer = true;
			hasInitializer = true;
			//System.out.println(node + " in " + cls);
		} else {
			isInInitializer = false;
		}
		return isInInitializer;
	}

	// flags say if this is a static block
	public boolean visit(Initializer node) {
		if (isInsideClass && Modifier.isStatic(node.getModifiers())) {
			isInInitializer = true;
			hasInitializer = true;
			//System.out.println(node + " in " + cls);
		} else {
			isInInitializer = false;
		}
		return isInInitializer;
	}

	public boolean visit(MethodDeclaration node) {
		if (isInInitializer) {
			super.visit(node);
		}
		return true;
	}

	public boolean visit(MethodInvocation node) {
		if (isInInitializer) {
			//System.out.println("Visiting... " + node);
			super.visit(node);
		}
		return true;
	}

	// this can be the name of a static field
	public boolean visit(SimpleName node) {
		if (isInInitializer) {
			super.visit(node);
		}
		return true;
	}

	// no way this can be in an initializer
	// so we don't continue the traversal down the tree
	public boolean visit(SuperConstructorInvocation node) {
		return false;
	}

	// no way this can be in an initializer
	// so we don't continue the traversal down the tree
	public boolean visit(SuperMethodInvocation node) {
		return false;
	}

	// this is to remember the parent
	// in case super was not called explicitly and
	// we have to call it manually
	public boolean visit(TypeDeclaration node) {
		ITypeBinding itb = node.resolveBinding();
		IWorklistType temp = InfoDictionary.v().getTypeForBinding(itb);
		if (cls.equals(temp)) {
			isInsideClass = true;
		} else if (isInsideClass) {
			return false;
		}
		// go down other types because the type we are searching for
		// may be inner
		return true; 
	}

	// set isInClass to false after desired class is visited
	public void endVisit(TypeDeclaration node) {
		ITypeBinding itb = node.resolveBinding();
		IWorklistType temp = InfoDictionary.v().getTypeForBinding(itb);
		if (cls.equals(temp)) {
			isInsideClass = false;
		}
	}

}
