/*
 * Created on Jul 18, 2005
 *
 */
package tacle.internal.rta.data;

import java.io.PrintWriter;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.HashMap;
import java.util.HashSet;

import org.eclipse.jdt.core.dom.ITypeBinding;

/**
 * @author Mariana L Barca
 * 
 * A collection of caches that speed up the execution of the worklist RTA
 * Singleton
 */
public class InfoDictionary {
	private Map classInfos; // Integer class id -> ClassInfo

	private Map typesByName; // String class name -> IWorklistClass

	private Map methodInfos; // Integer method id -> MethodInfo

	private Map methodInfosByClassId; // Integer class id -> Map

	// String methodName -> Set of MethodInfos

	private Map astMethods; // String key -> ASTMethod

	private Map otherTypes; // Integer type id -> ASTClass

	private Map packageIdByName; // String package name -> Integer package id

	private Map packageNameById; // Integer package id -> Stirng package name

	private Map typeKeys; // IWorklistClass type -> String key

	private Set loadedPackages; // Integer package id

	// private unique instance
	private static InfoDictionary instance;

	// private constructor
	private InfoDictionary() {
		classInfos = new HashMap();
		typesByName = new HashMap();
		methodInfos = new HashMap();
		methodInfosByClassId = new HashMap();
		astMethods = new HashMap();
		otherTypes = new HashMap();
		packageIdByName = new HashMap();
		packageNameById = new HashMap();
		typeKeys = new HashMap();
		loadedPackages = new HashSet();
		DataReader.v().loadAllOtherTypes(otherTypes, typesByName);
		DataReader.v().loadAllPackages(packageIdByName, packageNameById);
	}

	// returns the unique instance of this class
	public static InfoDictionary v() {
		if (instance == null) {
			instance = new InfoDictionary();
		}
		return instance;
	}

	// get the name of types that are not declared in a class
	// based on id
	public ASTType getOtherType(int typeId) {
		Object result = otherTypes.get(new Integer(typeId));
		if (result != null)
			return (ASTType) result;
		return null;
	}

	// get class info for class and package ids
	public ClassInfo getClassInfo(int classId, int pkgId) {
		Integer pkgIdInt = new Integer(pkgId);
		if (!loadedPackages.contains(pkgIdInt)) {
			loadPackage(pkgIdInt);
		}
		Object result = classInfos.get(new Integer(classId));
		if (result != null)
			return (ClassInfo) result;
		return null;
	}

	// get classInfo for name and package id
	// if it has packageid, must be a ClassInfo
	public ClassInfo getClassInfoForName(String className, int pkgId) {
		Integer pkgIdInt = new Integer(pkgId);
		if (!loadedPackages.contains(pkgIdInt)) {
			loadPackage(pkgIdInt);
		}
		Object result = typesByName.get(className);
		if (result != null)
			return (ClassInfo) result;
		return null;
	}

	// returns first IWorklistClass found for this type binding
	public IWorklistType getTypeForBinding(ITypeBinding tb) {
		IWorklistType result = null;
		String qName = tb.getQualifiedName();

		// search in typesByName for a class already loaded
		if (!qName.equals("")) {
			result = (IWorklistType) typesByName.get(qName);
			if (result != null) {
				setTypeKey(result, tb.getKey());
				return result;
			}
		}
		String bName = tb.getBinaryName();
		// search in typesByName for a weird class already loaded
		if (!tb.isTopLevel() && !tb.isPrimitive() && !tb.isArray()) {
			if (!bName.equals("")) {
				result = (IWorklistType) typesByName.get(bName);
				if (result != null) {
					setTypeKey(result, tb.getKey());
					return result;
				}
			}
		}
		// search in the summary on disk
		if (tb.getPackage() != null) {
			int packageId = getPackageIdForName(tb.getPackage().getName());
			if (packageId > 0) {
				if (!qName.equals("")) {
					result = getClassInfoForName(qName, packageId);
					if (result != null) {
						return result;
					}
				}
				if (!tb.isTopLevel() && !tb.isPrimitive() && !tb.isArray()
						&& !bName.equals("")) {
					result = getClassInfoForName(bName, packageId);
					if (result != null) {
						return result;
					}
				}
			}
		}
		// it's probably a new user class
		result = new ASTType(tb, false);
		// for safety, add alternate names for weird classes
		if (!qName.equals("") && !bName.equals("") && !qName.equals(bName)) {
			typesByName.put(qName, result);
			typesByName.put(bName, result);
		} else { // normal class
			typesByName.put(result.getCompleteName(), result);
		}
		// add type keys, for the search of ASTs
		setTypeKey(result, tb.getKey());
		return result;
	}

	// get type for name
	// shortcut method, to be applied only for "other" types when their names
	// are known
	public ASTType getOtherTypeForName(String typeName) {
		Object result = typesByName.get(typeName);
		if (result != null)
			return (ASTType) result;
		return null;
	}

	// we assume this is NOT a lib type
	// type will be added artificially, if it doesn't exist
	public ASTType addOtherTypeForName(String packageName, String typeName,
				String simpleName, String parentName, int position) {
		ASTType t = getOtherTypeForName(typeName);
		if (t==null) {
			t = new ASTType(packageName, typeName, simpleName, false);
			t.setPosition(position, parentName);
			typesByName.put(typeName, t);
			// class key is unknown, will be added later
		}
		return t;
	}
	
	public ASTType getMainClass(String packageName, String className) {
		// we assume the main class is always top level
		String simpleName = className;
		if (className.lastIndexOf('.')>-1)
			simpleName = className.substring(className.lastIndexOf('.'));
		ASTType m = new ASTType(packageName, className, simpleName, false);
		typesByName.put(className, m);
		// class key is unknown, will be added later
		return m;
	}

	public MethodInfo getMethodInfoForId(int methodId, int declaringpkgId) {
		Integer pkgIdInt = new Integer(declaringpkgId);
		if (!loadedPackages.contains(pkgIdInt)) {
			loadPackage(pkgIdInt);
		}
		Object result = methodInfos.get(new Integer(methodId));
		if (result != null)
			return (MethodInfo) result;
		return null;
	}

	public ASTMethod getASTMethod(String key) {
		Object result = astMethods.get(key);
		if (result != null)
			return (ASTMethod) result;
		return null;
	}

	public MethodInfo getClinitInfo(ClassInfo ci) {
		return getMethodInfoByName("<clinit>", ci, new IWorklistType[0]);
	}

	public int getPackageIdForName(String pkgName) {
		if (pkgName.equals(""))
			return 0;
		if (packageIdByName.containsKey(pkgName)) {
			return ((Integer) packageIdByName.get(pkgName)).intValue();
		}
		return 0;
	}

	public String getPackageNameForId(int pkgId) {
		Integer pkgIdInt = new Integer(pkgId);
		if (packageNameById.containsKey(pkgIdInt)) {
			return (String) packageNameById.get(pkgIdInt);
		}
		return "";
	}

	public Map getMethodsByClass(ClassInfo ci) {
		Integer packageId = new Integer(ci.getPackageId());
		Integer classId = new Integer(ci.getClassId());
		if (!loadedPackages.contains(packageId)) {
			loadPackage(packageId);
		}
		return (Map) methodInfosByClassId.get(classId);

	}

	public MethodInfo getMethodInfoByName(String name, IWorklistType declClass,
			IWorklistType[] paramTypes) {
		if (declClass instanceof ClassInfo) {
			// this method triggers the loading of declPkg if not loaded already
			ClassInfo declClassInfo = (ClassInfo) declClass;
			int declClassId = declClassInfo.getClassId();
			Map methodsByName = (Map) methodInfosByClassId.get(new Integer(
					declClassId));
			if (methodsByName != null) {
				Set methods = (Set) methodsByName.get(name);
				if (methods != null) {
					for (Iterator it = methods.iterator(); it.hasNext();) {
						MethodInfo mi = (MethodInfo) it.next();
						// check signature
						IWorklistType[] params = mi.getParameters();
						if (params.length == paramTypes.length) {
							boolean sameParams = true;
							for (int i = 0; i < params.length; i++) {
								
								// test
								if (paramTypes[i]==null) {
									System.out.println(name + " in " + declClass.getCompleteName());
								}
								if (!paramTypes[i].equals(params[i]))
									sameParams = false;
							}
							if (sameParams)
								return mi;
						}
					}
				}
			}
		}
		return null;
	}

	public void setMethodInfo(int methodId, MethodInfo mi) {
		methodInfos.put(new Integer(methodId), mi);
	}

	public void setASTMethod(String key, ASTMethod m) {
		astMethods.put(key, m);
	}

	public void printClassMap() {
		for (Iterator it = typesByName.keySet().iterator(); it.hasNext();) {
			Object key = it.next();
			System.out.print("String " + key);
			IWorklistType ci = (IWorklistType) typesByName.get(key);
			System.out.println(" -> " + ci.getCompleteName());
		}
	}

	// search among non class types a type with a certain name
	public int getOtherTypeId(ASTType otherType) {
		if (otherTypes.containsValue(otherType)) {
			for (Iterator it = otherTypes.keySet().iterator(); it.hasNext();) {
				Integer n = (Integer) it.next();
				if (otherTypes.get(n).equals(otherType)) {
					return n.intValue();
				}
			}
		}
		return 0;
	}

	private void setTypeKey(IWorklistType type, String typeKey) {
		typeKeys.put(type, typeKey);
	}

	// used only for AST types
	public String getTypeKey(IWorklistType type) {
		return (String) typeKeys.get(type);
	}

	private void loadPackage(Integer packageId) {
		loadedPackages.add(packageId);
		// load classes
		DataReader.v().loadAllClasses(packageId, classInfos, typesByName);
		// load methods
		DataReader.v().loadAllMethods(packageId, methodInfos,
				methodInfosByClassId);
		// load init classes
		DataReader.v().loadAllMethodInitializesClass(packageId, methodInfos);
		// load inst classes
		DataReader.v().loadAllMethodInstantiatesClass(packageId, methodInfos);
		// load call sites
		DataReader.v().loadAllMethodHasCallSite(packageId, methodInfos);
	}

	public void printStat(PrintWriter pw) {
		// print methodInfosByClassId // Integer -> Map // String methodName ->
		// Set of MethodInfos
		for (Iterator it = methodInfosByClassId.keySet().iterator(); it
				.hasNext();) {
			Object key = it.next();
			pw.println(key);
			Map map = (Map) methodInfosByClassId.get(key);
			if (map != null) {
				for (Iterator it1 = map.keySet().iterator(); it1.hasNext();) {
					Object key1 = it1.next();
					pw.println(" -> " + key1);
					Set methods = (Set) map.get(key1);
					if (methods != null) {
						for (Iterator it2 = methods.iterator(); it2.hasNext();) {
							pw.println("    -> " + it2.next());
						}
					}
				}
			}
		}
	}
}
